#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals 		= 3
#pragma IgorVersion		= 6.36
#pragma version			= 3.71
#pragma DefaultTab		= {3,20,4}		// Set default tab width in Igor Pro 9 and later

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21549																// the project node on IgorExchange
static StrConstant ksShortTitle = "Profile Generator"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//
//	SpectraProfile is a environment to create 1D profiles from 2D data sets in horizontal or vertical
//	direction. Various modifications to the 2D data itself can be applied. Some functionality relies on
//	properties of the data used (such as scaling and column labels) and can be changed below.
//	
// 2021-02-16 - ver. 3.60:	Initial public release.
// 2021-07-22 - ver. 3.70:	Slightly adjusted the names in the main menu entries.
// 							Fixed wrong wave labels (row vs. col) for the split tool.
//							Added scaled positions to the wave name for the split tool.
//							Added keyboard shortcuts for Normalize Profile and Snap to Grid.
//							Added Accept & Quit button.
//							Added optional logging of data edits in the history.
//							Added control to set the smoothing factor for the align feature.
//							Moved all controls to an external panel.
//							Now the live profile under the mouse cursor gets normalized as well.
//							Fixed the way-too-big smoothing value when using the align feature.
//							Data units get displayed in the graph now.
// 2021-07-24 - ver. 3.71:	Improvements of matrix calculations thanks to suggestions by A.G. (WaveMetrics).
//							Added image contrast controls.
//________________________________________________________________________________________________

StrConstant	kProfileWorkingDir	= "Profile_Generator"											// the name for the working folder
StrConstant	kProfileTitle		= "Spectra Image Profile (ver. 3.71 - 07/2021): Profile of "	// string for the title bar of the main window
Constant	kNormProfileSmooth	= 2																// profile normalization: integer factor for smoothing the profile before division (0 = no smoothing)
Constant	kAlignSmooth		= 2																// align features: smoothing factor in percent of the wave's x/y dimension (e.g., 2 => points * 0.02)

//-------------------------------- functions for user code ---------------------------------------
Static StrConstant kUserInfoBtnName = "Ring Current"											// the title of the user button

Static Function Profile_ActivateUserInfoButton(inwave)											// called to check whether or not the data fulfills conditions for measurement normalization
	Wave inwave
	Variable activate = 0
	// +++++++++++++++++++++++++++ conditions for activation +++++++++++++++++++++++++++++++++++++
	activate = StringMatch(GetDimLabel(inwave,1,0),"*mA*")										// check if dim-labels contain "mA" values
	activate = StringMatch(note(inwave),"*ring current normalized*")? 0 : activate				// already normalized?
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return activate																				// return 1 to activate button
End

Static Function Profile_ExecuteUserInfoButton(inwave)											// normalization of the data to measurement info (when the button is pressed)
	Wave inwave
	Variable valAvg = 0
	// ++++++++++++++++++++++ ring current normalization procedure +++++++++++++++++++++++++++++++
	Variable i, value
	for (i = 0; i < DimSize(inwave,1); i += 1)
		String Currlabel = GetDimLabel(inwave,1,i)												// get the information from the column label
		sscanf Currlabel, "%s (%f mA)", Currlabel, value										// search for ring current info in the form "time (XXX mA)"
		if (value != 0)
			inwave[][i] /= value
		endif
		valAvg += value																			// add the values for intensity re-normalization later
	endfor
	valAvg /= i+1
	Note inwave, "ring current normalized"														// write into note that the normalization has been done
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return valAvg																				// return the intensity normalization factor which may be used restore the average height
End
//------------------------------------------------------------------------------------------------

Menu "Spectra Tools"
	"Image Profile ..."						,/Q, CreateBrowser; SpectraProfileWrapper()
	"-"
	"Append 1D,2D Datasets into 2D Data"	,/Q, CreateBrowser; CombineDataSets()
	"Stack 2D Datasets into 3D Data"		,/Q, CreateBrowser; CombineDataSets(stack3D=1)
	"-"
	"Profile of 2D|3D Data - [Horizontal]"	,/Q, CreateBrowser; QuickProfileWrapper(0)
	"Profile of 2D|3D Data - [Vertical]"	,/Q, CreateBrowser; QuickProfileWrapper(1)
	"2D Image from 3D Data - [Layers]"		,/Q, CreateBrowser; QuickProfileWrapper(2)
	"-"
	"Split Up 2D|3D Data - [Horizontal]"	,/Q, CreateBrowser; QImageSplitWrapper(0)
	"Split Up 2D|3D Data - [Vertical]"		,/Q, CreateBrowser; QImageSplitWrapper(1)
	"Split Up 3D Data - [Layers]"			,/Q, CreateBrowser; QImageSplitWrapper(2)
End

//________________________________________________________________________________________________
//	Quick Profile - sums 2D/3D waves in one direction (along x, y or z) for data reduction
//________________________________________________________________________________________________

Function QuickProfileWrapper(dim)
	Variable dim
	Variable index=0
	do
		String name = GetBrowserSelection(index)
		if (strlen(name) <= 0)
			break
		endif
		Wave work = $name
		QProfile(work,dim)	
		index+=1
	while(1)
	return 0
End
//################################################################################################
Function QProfile(inwave,dim)
	Wave inwave
	Variable dim 		// 0 = H, 1 = V, 2 = L
	
	Variable wdim = WaveDims(inwave)
	String name = NameOfWave(inwave)
	
	dim = round(dim)
	if (dim < 0 || dim > 2)
		Abort "Wrong dimension parameter. Must be 0, 1 or 2"
	endif
	if (wdim < 2 && dim == 1)
		Abort "This is meant for 2D waves!"
	endif
	if (wdim != 3 && dim == 2)
		Abort "This is meant for 3D waves!"
	endif

	Variable chars = IgorVersion() < 7.00 ? 26 : inf
	String outName = name[0,chars] + SelectString(dim-1, "_prH", "_prV","_prL")					// the name of the output wave
	if (dim==0)
		MatrixOP/O $outName = sumRows(inwave)/numCols(inwave)
	elseif (dim==1)
		MatrixOP/O $outName = sumRows(inwave^T)/numRows(inwave)
	elseif (dim==2)
		MatrixOP/O $outName = sumBeams(inwave)/waveLayers(inwave)
	endif
	Wave outwave = $outName
	
	if (wdim == 3 && dim < 2)																	// for profiles of 3D waves -> rotate result to 2D map
		MatrixOP/FREE temp = transposeVol(outwave,1)
		Duplicate/O temp, outwave
		SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), outwave
	endif
	
	String notes = note(inwave) + "\r" + name													// scale & write a note of what has been done
	if (dim < 2)
		SetScale/P x, DimOffset(inwave,dim), DimDelta(inwave,dim), WaveUnits(inwave,dim), outwave
		notes += " integrated from "+num2str(DimOffset(inwave,1-dim))+" to "+num2str(DimOffset(inwave,1-dim)+(DimSize(inwave,1-dim)-1)*DimDelta(inwave,1-dim))+" "+WaveUnits(inwave,1-dim)
	elseif (dim == 2)
		CopyScales inwave, outwave
		notes += " integrated over all layers."
	endif
	Note outwave, notes
	
	return 0
End

//________________________________________________________________________________________________
//	Image split - splits 2D/3D data into individual waves along x, y or z
//________________________________________________________________________________________________

Function QImageSplitWrapper(dim)
	Variable dim
	Variable index=0
	do
		String name = GetBrowserSelection(index)
		if (strlen(name) <= 0)
			break
		endif
		Wave work = $name
		QImageSplit(work,dim)	
		index+=1
	while(1)
	return 0
End
//################################################################################################
Function QImageSplit(inwave,dim)
	Wave inwave
	Variable dim 					// 0 = H, 1 = V, 2 = L
	
	Variable wdim = WaveDims(inwave)
	String name = NameOfWave(inwave)
	
	dim = round(dim)
	if (dim < 0 || dim > 2)
		Abort "Wrong dimension parameter. Must be 0, 1 or 2"
	endif
	if (wdim < 2 && dim == 1)
		Abort "This is meant for 2D or 3D waves!"
	endif
	if (wdim != 3 && dim == 2)
		Abort "This is meant for 3D waves!"
	endif
	
	DFREF saveDFR = GetDataFolderDFR()
	NewDataFolder/O/S $name
	
	Variable i
	String wname, unit, scaledpos
	Switch (dim)
		case 0:
			unit = WaveUnits(inwave,1)
			for (i = 0; i < DimSize(inwave, 1); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,1)+i*DimDelta(inwave,1)) + SelectString(strlen(unit),""," ") + unit + ")"
				if (wdim == 3)
					wname = "col_img_" + num2str(i) + scaledpos
					MatrixOP/Free temp = col(inWave,i)
					MatrixOP/O $wname = transposeVol(temp,1)
					SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
					SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), $wname
				else
					wname = "col_" + num2str(i) + scaledpos
					MatrixOP/O $wname = col(inWave,i)
					SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
				endif
			endfor
		break
		case 1:
			unit = WaveUnits(inwave,0)
			for (i = 0; i < DimSize(inwave, 0); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,0)+i*DimDelta(inwave,0)) + SelectString(strlen(unit),""," ") + unit + ")"
				if (wdim == 3)
					wname = "row_img_" + num2str(i) + scaledpos
					MatrixOP/Free temp = row(inWave,i)
					MatrixOP/O $wname = transposeVol(temp,4)
					SetScale/P x, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
					SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), $wname
				else
					wname = "row_" + num2str(i) + scaledpos
					MatrixOP/O $wname = row(inWave,i)^t
					SetScale/P x, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
				endif
			endfor
		break
		case 2:
			unit = WaveUnits(inwave,2)
			for (i = 0; i < DimSize(inwave, 2); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,2)+i*DimDelta(inwave,2)) + SelectString(strlen(unit),""," ") + unit + ")"
				wname = "layer_" + num2str(i) + scaledpos
				MatrixOP/O $wname=inWave[][][i]
				SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
				SetScale/P y, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
			endfor
		break
	EndSwitch
	SetDataFolder saveDFR
	
	return 0
End

//________________________________________________________________________________________________
//	Combines individual 1D or 2D datasets to one 2D/3D dataset
//________________________________________________________________________________________________

Function CombineDataSets([stack3D])
	Variable stack3D																			// stack3D = 1 concatenates the data without summing
	
	String name = "", list = ""
	Variable index = 0, items
	do																							// make a list of the selected data
		name = GetBrowserSelection(index)
		if (strlen(name) <= 0)
			break
		endif
		list += name + ";"
		index+=1
	while(1)
	list  = Sortlist(list, ";", 16)
	items = itemsinlist(list)
	
	// +++++++++++++++++++++++++++++++++ build an output name ++++++++++++++++++++++++++++++++++++
	String number1, number2, outname
	name = ParseFilePath(0, StringFromList(items-1, list), ":", 1, 0)							// try to get the last data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number2													// get the end number in the name
	name = ParseFilePath(0, StringFromList(0, list), ":", 1, 0)									// try to get the first data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number1
	
	outname = RemoveEnding(ReplaceString("'",name,""), number1)
	if(strlen(number1) > 0 && strlen (number2) > 0)												// add range to name if both numbers are set
		outname += "_"+number1+"-"+number2
	endif
	Variable chars = IgorVersion() < 7.00 ? 26 : inf
	outname = outname[0,chars] + "_cmb"
	
	if (!ParamIsDefault(stack3D) && stack3D == 1)
		Concatenate/O list, $outname
		return 0																				// stop here
	endif
	
	// ++++++++++++++++++++++++++++ continue when stack3D is not set +++++++++++++++++++++++++++++
	Make/FREE/N=(items) xSizes, ySizes, Offset, Deltas, RngEnd									// temporary data holding all the scaling and unit values from all waves
	Make/FREE/N=(items)/T xUnits, yUnits
	
	Wave work 		= $StringFromList(0, list)													// extract y-scaling and names from first wave
	Variable yOffset= DimOffset(work, 1)
	Variable yDelta	= DimDelta(work, 1)
	String yUnit	= WaveUnits(work, 1)
	String wname	= NameOfWave(work)
	String notes	= note(work)
	
	For (index = 0; index < items; index += 1)
		Wave work = $StringFromList(index, list)
		xSizes[index] = DimSize(work, 0)
		ySizes[index] = DimSize(work, 1) > 0 ? DimSize(work, 1) : 1
		Offset[index] = DimOffset(work, 0)
		Deltas[index] = DimDelta(work, 0)
		RngEnd[index] = Offset[index] + Deltas[index] * (xSizes[index]-1)
		xUnits[index] = WaveUnits(work, 0)
	Endfor
	
	Variable Delta = Deltas[0], DeltaMisMatch = 0
	if (Delta != mean(Deltas))																	// may lead to data overrun
		DoAlert 1, "Data x spacing does not match! Do you want to use interpolation? Press no to exit."
		if (V_flag == 1)
			Delta = Delta > 0 ? WaveMin(Deltas) : WaveMax(Deltas)
			DeltaMisMatch = 1
		else
			return 0
		endif
	endif
	
	Variable WaveStart	= Delta > 0? WaveMin(Offset) : WaveMax(Offset)
	Variable WaveEnd	= Delta > 0? WaveMax(RngEnd) : WaveMin(RngEnd)
	Variable xSize		= WaveMax(xSizes)
	if (DeltaMisMatch)
		xSize = abs((WaveStart-WaveEnd)/Delta)+1
	endif
	
	Make/O/D/N=(xSize, Sum(Ysizes)) $outname = 0
	Wave out = $outname
	SetScale/I x, WaveStart, WaveEnd, xUnits[0], out
	SetScale/P y, yOffset, yDelta, yUnit, out
	Note out, notes

	Variable startpos = 0, endpos, startpnt, endpnt
	For (index = 0; index < itemsinlist(list); index += 1)
		endpos	  = startpos + ySizes[index]-1
		WaveStart = Offset[index]
		Wave work = $StringFromList(index, list)

		startpnt  = x2pnt(out, pnt2x(work, 0))													// find the sub-range of valid x-values from work
		endpnt	  = x2pnt(out, pnt2x(work, xSizes[index]-1))
		startpnt += (pnt2x(work, 0) - pnt2x(out, startpnt) > 1e-6)								// make sure the x-range is not even slightly outside the valid range
		endpnt	 -= (pnt2x(out, endpnt) - pnt2x(work, xSizes[index]-1) > 1e-6)	

		out[startpnt,endpnt][startpos,endpos] = work(x)[q-startpos] 
		startpos	+= ySizes[index]
	Endfor
	
	return 0
End

//________________________________________________________________________________________________
//	Generate profiles from 2D waves
//________________________________________________________________________________________________

Function SpectraProfileWrapper()
	Wave/Z work = $GetBrowserSelection(0)	// first selection
	if (WaveExists(work))
		SpectraProfile(work)
	endif
	return 0
End

//################################################################################################

Function SpectraProfile(inwave)
	Wave inwave
	if (WaveDims(inwave) != 2)
		Abort "This is meant for 2D waves!"
	endif
	if (WinType("ProfileGeneratorWindow"))
		DoWindow/F ProfileGeneratorWindow
		DoAlert/T="Open Image Profile window found." 1, "There is another instance open. Close the old one and load the new data? Any changes will not be saved!"
		if (V_flag != 1)
			return 0
		endif
		Profile_QuitProgram(1)																	// quit old panel and undo all changes
	endif

	NewDataFolder/O	root:Packages
	NewDataFolder/O	root:Packages:$(kProfileWorkingDir)											// the folder for temporary stuff
	DFREF storage = root:Packages:$(kProfileWorkingDir)
	
	String/G		storage:InwaveLocation = GetWavesDataFolder(inwave,2)						// save wave name and folder for later
	Variable/G		storage:ProfileUpdateConnector												// formula connection variable... has no purpose itself, but is an anchor for triggering updates
	Make/O/D 		storage:SectionLine															// the currently displayed profile
	Make/O/D/N=5	storage:CurrentLines = 0													// holds the range values ([0] = current working mode {0 = horizontal, 1 = vertical}, [1] = h.pos, [2] = h.width, [3] = v.pos, [4] = v.width)
	Make/O/D/N=5 	storage:LinesY, storage:LinesX												// marquee line waves for the image 
	SetScale d 0,0,WaveUnits(inwave,0), storage:LinesX											// the data units will be displayed on the x- and y-axes
	SetScale d 0,0,WaveUnits(inwave,1), storage:LinesY
	Duplicate/O		inwave, storage:StartCopy													// copy original wave for undo

	NewDataFolder/O	root:WinGlobals
	NewDataFolder/O	root:WinGlobals:ProfileGeneratorGraph										// here all main window related data is generated
	String/G		root:WinGlobals:ProfileGeneratorGraph:S_TraceOffsetInfo = ""				// Igor defined window status parameter
	//+++++++++++++++++++++++++++++++++ generate the panels ++++++++++++++++++++++++++++++++++++++
	Wave LinesX = storage:LinesX
	Wave LinesY	= storage:LinesY
	Variable pix = 72/ScreenResolution
	Variable TFTResX= Str2Num(StringFromList(3,StringByKey("SCREEN1",IgorInfo(0),":"),","))		// get the display resolution to align window better (center of full screen)
	Variable TFTResY= Str2Num(StringFromList(4,StringByKey("SCREEN1",IgorInfo(0),":"),","))
	Variable WinTop = 70, WinLeft = 200
	Variable WinBottom	= TFTResY - 150
	Variable WinRight	= TFTResX - 350

	Display/K=2/W=(WinLeft*pix, WinTop*pix, WinRight*pix, WinBottom*pix)/N=ProfileGeneratorGraph as kProfileTitle+NameOfWave(inwave)
	AppendImage inwave
	ModifyImage ''#0 ctab= {*,*,Terrain,0}
	AppendToGraph/C=(0,0,0) LinesY vs LinesX													// add marquee lines
	ModifyGraph quickdrag(LinesY)=1,live(LinesY)=1												// lines can be dragged (act on S_TraceOffsetInfo)
	
	SetWindow ProfileGeneratorGraph hook(CreateProfile)=Profile_ProcessWindowEvent				// keeps track of window changes
	NVar ProfileUpdateConnector = storage:ProfileUpdateConnector								// set formula for live update
	SetFormula ProfileUpdateConnector,"LinesIntegrateUpdate(root:WinGlobals:ProfileGeneratorGraph:S_TraceOffsetInfo)"
	//++++++++++++++++++++++++++++++++++++ now the controls ++++++++++++++++++++++++++++++++++++++
	NewPanel/HOST=ProfileGeneratorGraph/EXT=0/K=2/N=ProfileGeneratorWindow/W=(0, 0, 270, 563) as "Spectra Profile"
	SetWindow ProfileGeneratorGraph#ProfileGeneratorWindow hook(CreateProfile)=Profile_ProcessWindowEvent	// keeps track of window changes here as well (only button presses)
	
	TitleBox IntMode			,pos={10,10}	,size={80,23}	,title="Mode:"	 			,frame=0	,fstyle=1	,fSize=12
	CheckBox IntColCheckBox		,pos={60,10}	,size={80,23}	,title="\f04H\f00orizontal: ="		,help={"Select horizontal mode. (key: 'h')"}
	CheckBox IntRowCheckBox		,pos={165,10}	,size={80,23}	,title="\f04V\f00ertical: ||"		,help={"Select vertical mode. (key: 'v')"}
	SetVariable Position		,pos={10,33}	,size={150,23}	,title="Center:"					,help={"Center position of the selection. Move with cursor keys."}	
	SetVariable Width			,pos={10,58}	,size={150,23}	,title="Width:"						,help={"Width of the selection. This is the distance between the two lines. Move with cursor keys."}
	CheckBox FullWidth			,pos={170,34}	,size={70,23}	,title="Full Size"					,help={"Select the full range of the spectrum. (key: page up)"}
	CheckBox MinWidth			,pos={170,59}	,size={70,23}	,title="Minimal Size"				,help={"Select one slice. (key: page down)"}
	
	Button bSaveSpec			,pos={20,90}	,size={100,26}	,title="\f04S\f00ave Profile"		,help={"Save current profile. (key: 's') Press 'shift' to omit range in the wave name."}
	Button bCompare				,pos={20,125}	,size={100,26}	,title="Sna\f04p\f00shot"			,help={"Click to append and compare current profile. (key: 'p')"}
	
	CheckBox Normto1			,pos={140,90}	,size={80,23}	,title="\f04N\f00ormalize Profile"	,help={"Normalizes the profile to 1. (key: 'n')"}
	CheckBox MousetoSpec		,pos={140,112}	,size={80,23}	,title="\f04M\f00ouse Csr Profile"	,help={"Live profile under the mouse cursor. (key: 'm')"}
	CheckBox SnapToGrid			,pos={140,134}	,size={80,23}	,title="Snap Lines to \f04G\f00rid"	,help={"Snap the selection lines to full rows/columns. (key: 'g')"}

	CheckBox cPrintEdits		,pos={22,161}	,size={250,23}	,title="Print a Log of Data Edits to History"	,fstyle=2

	GroupBox NormalizeGroup		,pos={10,185}	,size={250,83}	,title="Normalize 2D Data"	,frame=0	,fstyle=1
	CheckBox MaintainInt		,pos={22,210}	,size={80,23}	,title="Keep average Int."			,help={"For normalization: Check to maintain the (average) height of the spectrum. Otherwise the spectrum will be scaled down while normalizing."}	
	Button bNormalize			,pos={20,235}	,size={110,23}	,title="Div. by Profile"			,help={"Normalize the spectrum with the current profile (divide). Holding 'ctrl / command' disables smoothing (hard normalize)."}
	Button bNormUser			,pos={140,205}	,size={110,23}	,title=kUserInfoBtnName				,help={"Normalize the spectrum with the included measurement information, e.g., ring current."}
	Button bNormMax				,pos={140,235}	,size={110,23}	,title="Each Maximum"				,help={"Normalize to the maximum of every horizontal / vertical line."}
	
	GroupBox ModifyGroup		,pos={10,275}	,size={250,213}	,title="Edit 2D Data"		,frame=0	,fstyle=1
	Button bSubtractPr			,pos={20,300}	,size={130,23}	,title="Subtract Profile"			,help={"Cut the current Profile from the spectrum. Press and hold 'ctrl / command' (or 'alt / option') to subtract along features moving with increasing (decreasing) Y values."}
	Button bRemoveSelection		,pos={20,330}	,size={130,23}	,title="Delete ROI Section"			,help={"Remove the selected data between the lines. (key: del)"}
	Button bShiftYAxisPlus		,pos={160,300}	,size={90,23}	,title="> Shift: +Y"				,help={"Shifts data by adding Y-axis values to X values. Hold the 'ctrl / command' key, to cut image to exclude partial data."}
	Button bShiftYAxisMinus		,pos={160,330}	,size={90,23}	,title="< Shift: -Y"				,help={"Shifts data by subtracting Y-axis values from X values. Hold the 'ctrl / command' key, to cut image to exclude partial data."}

	GroupBox SeparationLine1	,pos={20,359}	,size={228,1}	,frame=0
	Button bAlign				,pos={20,368}	,size={80,40}	,title="Align\nFeatures"			,help={"Use data correlation to align features in the image (press 'shift' for top to bottom alignment). Set cursors to align to a sub-region!"}
	CheckBox cAlignSmooth		,pos={110,370}	,size={80,23}	,title="Smoothing:"					,help={"Smooth profile before trying alignment (helps with low S/N)."}
	CheckBox cAlignOutput		,pos={110,391}	,size={80,23}	,title="Output Shift Result"		,help={"Saves result of alignment in a wave."}
	SetVariable SmoothFactor	,pos={140,368}	,size={110,23}	,title=""							,help={"The smoothing factor (no. of binomial smoothing operations) applied before alignment."}

	GroupBox SeparationLine2	,pos={20,415}	,size={228,1}	,frame=0
	Button bUndoAll				,pos={20,425}	,size={150,23}	,title="\f05U\f01ndo All Changes"	,help={"Restore the original wave. (key: 'u')"} ,fSize=13
	Button bUndoLast			,pos={20,455}	,size={150,23}	,title="Undo Last Change"			,help={"Restore the data before the last action."}
	Button bSaveQuit			,pos={180,425}	,size={70,53}	,title="Accept\r& Quit"				,help={"Keeps any changes made to the 2D data and closes the window."}
	
	TitleBox ContrastLabel		,pos={10,497}	,size={80,23}	,title="Image Contrast:"	,frame=0	,fStyle=1
	TitleBox ContrLowLabel		,pos={10,517}	,size={24,23}	,title="Min:"				,frame=0
	TitleBox ContrHighLabel		,pos={10,537}	,size={24,23}	,title="Max:"				,frame=0
	
	PopupMenu ContrastColor		,pos={120,495}	,size={140,20}	,title=""	,value= "*COLORTABLEPOPNONAMES*"
	Slider ContrastLow			,pos={48,522}	,size={210,10}	,side=0	,vert=0	,ticks=0
	Slider ContrastHigh			,pos={48,542}	,size={210,10}	,side=0	,vert=0	,ticks=0
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList("ProfileGeneratorGraph#ProfileGeneratorWindow",";","*") fsize=10
	endif
	
	ModifyControlList ControlNameList("ProfileGeneratorGraph#ProfileGeneratorWindow", ";", "b*")  proc=Profile_MainButtonProc	// define the procedure for all buttons
	Button bNormUser			,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)			// activate measurement info normalization if the conditions are right
	Button bSaveSpec			,fColor=(65535,0,0)
	CheckBox IntColCheckBox		,value = 1 ,mode = 1	,proc=ModeCheckBoxControl				// radio button checked => horizontal mode
	CheckBox IntRowCheckBox		,value = 0 ,mode = 1	,proc=ModeCheckBoxControl				// radio button unchecked
	CheckBox FullWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MinWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox Normto1			,value = 0				,proc=ModeCheckBoxControl
	CheckBox SnapToGrid			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MaintainInt		,value = 0
	CheckBox cAlignSmooth		,value = 0
	CheckBox cAlignOutput		,value = 0
	SetVariable Position		,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable Width			,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable SmoothFactor	,bodywidth = 55	,limits={0,32767,1}	,value=_NUM:(ceil(DimSize(inwave,0)*kAlignSmooth/100))
	PopupMenu ContrastColor		,bodyWidth = 140		,mode=1+WhichListItem("Terrain", CTabList()),proc=Profile_ContrastColorProc
	Slider ContrastLow			,value=WaveMin(inwave)	,limits={WaveMin(inwave),WaveMax(inwave),0}	,proc=Profile_ContrastSliderProc
	Slider ContrastHigh			,value=WaveMax(inwave)	,limits={WaveMin(inwave),WaveMax(inwave),0}	,proc=Profile_ContrastSliderProc
	//+++++++++++++++++++++++++++++++++++++ set dependency +++++++++++++++++++++++++++++++++++++++
	Profile_AppendSectionTrace()																// add the ROI lines & set variable limits for the first time
	Profile_SetMinFullLineValues(1,0,0)															// set initial values for row mode (position & width)
	Profile_SetMinFullLineValues(1,1,0)															// set initial values for col mode

	Profile_UpdateLines()																		// first update to generate initial profile
	Profile_SetFullMinCheckbox()
	DoWindow/F ProfileGeneratorGraph
	return 0
End

//################################################################################################

static Function/S getPanel()
	String panelStr = "ProfileGeneratorGraph#ProfileGeneratorWindow"
	if (WinType("ProfileGeneratorGraph") != 1)													// compatibility with older versions
		panelStr = "ProfileGeneratorWindow"
	endif
	return panelStr
End

//################################################################################################

static Function/S getGraph()
	String graphStr = "ProfileGeneratorGraph"
	if (WinType(graphStr) != 1)																	// compatibility with older versions
		graphStr = "ProfileGeneratorWindow"
	endif
	return graphStr
End

//################################################################################################

Function Profile_QuitProgram(Undo)
	Variable Undo
	DFREF stg = root:Packages:$(kProfileWorkingDir)
	DFREF pkg = root:Packages:
	
	if (WinType("ProfileGeneratorGraph"))
		KillWindow ProfileGeneratorGraph														// first, close image window
	elseif (WinType("ProfileGeneratorWindow"))
		KillWindow ProfileGeneratorWindow
	else
		return -1
	endif
	
	NVar/Z ProfileUpdateConnector = stg:ProfileUpdateConnector
	if (!NVAR_Exists(ProfileUpdateConnector))
		return -1																				// obviously the folder does not exist anymore
	endif
	SetFormula ProfileUpdateConnector,""														// reset formula
	
	if (Undo)																					// for some strange hook double call reason, the undo has to take place AFTER window reset
		SVAR WLoc = stg:InwaveLocation															// reference the modified wave
		Wave inwave = $WLoc
		Duplicate/O stg:StartCopy, inwave														// undo the changes before quitting
	endif
	
	KillDataFolder/Z stg																		// remove folders
	KillDataFolder/Z root:WinGlobals
	Variable pkgObjects = CountObjectsDFR(pkg, 1)+CountObjectsDFR(pkg, 2)+CountObjectsDFR(pkg, 3)+CountObjectsDFR(pkg, 4)
	if (pkgObjects == 0)																		// clean up Packages folder if empty
		KillDataFolder pkg
	endif
	return 0
End

//################################################################################################

Function Profile_AppendSectionTrace()															// appends the profile parallel to the selection lines
	DFREF stg = root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	Wave SecL	= stg:SectionLine
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable Mode = Lines[0]
	
	String unitStr = WaveUnits(inwave,1-Mode)
	Variable delta = DimDelta(inwave,1-Mode)
	Variable init  = abs(delta)*round((DimSize(inwave,1-Mode)-1) / 20)							// calculate ca. 5% of width
	init = init == 0 ? abs(delta) : init														// make sure that's not zero
	SetVariable Position	,win=$getPanel() ,limits={-inf,inf,init}		,value=Lines[1+Mode*2]	,format="%g "+unitStr	// set new control values
	SetVariable Width		,win=$getPanel() ,limits={abs(delta),inf,init}	,value=Lines[2+Mode*2]	,format="%g "+unitStr
	
	RemoveFromGraph/W=$getGraph()/Z CompareLine, SectionLine, MouseProfile						// remove all ROI traces
	If (Mode)	// vertical
		AppendToGraph/W=$getGraph()/VERT/B=bottom2 SecL
		ModifyGraph/W=$getGraph() axisEnab(left)={0,1}, axisEnab(bottom)={0.32,1}				// modify the main axes
		ModifyGraph/W=$getGraph() axisEnab(bottom2)={0,0.29}, freePos(bottom2)=0				// align the new bottom axis
		SetAxis/W=$getGraph()/A/E=1 bottom2														// scale from zero
	else			// horizontal
		AppendToGraph/W=$getGraph()/L=left2 SecL
		ModifyGraph/W=$getGraph() axisEnab(left)={0.52,1},axisEnab(bottom)={0,1}
		ModifyGraph/W=$getGraph() axisEnab(left2)={0,0.49}, freePos(left2)=0
		SetAxis/W=$getGraph()/A/E=1 left2
	endif
	ModifyGraph/W=$getGraph() highTrip=100000
	ModifyGraph/W=$getGraph() mirror=1, minor=1, standoff=0
	ModifyGraph/W=$getGraph() ZisZ=1, notation=1, rgb=(65280,0,0)
	return 0
End

//################################################################################################

Function Profile_ProcessWindowEvent(s)															// main window hook
	STRUCT WMWinHookStruct &s

	Variable HookTakeover = 0
	if (WinType(s.WinName) == 7 && s.EventCode != 11)											// process only key presses for the panel
		return HookTakeover
	endif

	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable Mode = Lines[0]
	
	Switch (s.EventCode)
		case 17:	// window close button
			HookTakeover = 1
			if (!EqualWaves(stg:startCopy, inwave,-1))											// compare if the wave was changed
				if (!Wintype("#AskQuitPanel"))													// this will create a little panel in case of changes
					Variable pix = 72/ScreenResolution
					Variable xCenter = (s.WinRect.right - s.WinRect.left)/2, yCenter = (s.WinRect.bottom - s.WinRect.top)/2
					NewPanel/N=AskQuitPanel/HOST=$s.WinName/W=((xCenter-112)*pix,(yCenter-35)*pix,(xCenter+112)*pix,(yCenter+35)*pix)
					DrawText 40,28,"Accept Changes to Data?"
					Button bAskOK		,pos={12,33}	,size={60,23}		,title="Yes"	,proc=Profile_AskButton
					Button bAskUndo		,pos={82,33}	,size={60,23}		,title="Undo"	,proc=Profile_AskButton
					Button bAskCancel	,pos={152,33}	,size={60,23}		,title="Cancel"	,proc=Profile_AskButton
					SetActiveSubwindow $s.WinName												// restore focus to main window
				endif
			else
				 Profile_QuitProgram(0)															// else just quit (no undo)
			endif
			break
		case 11:	// key press
			HookTakeover = 1
			//++++++++++++++++++++++++++++++++ execute button keys ++++++++++++++++++++++++++++++
			String Key = ""
			Switch(s.keycode)
				case 112:	// p key
					Key = "bCompare"
					break
				case 115:	// s key
					Key = "bSaveSpec"
					break
				case 117:	// u key
					Key = "bUndoAll"
					break
				case 127:	// del key
					Key = "bRemoveSelection"
					break
			EndSwitch
			if (StrLen(Key) > 0)
				Profile_ExecuteButtonCommand(Key)
				break
			endif
			//++++++++++++++++++++++++++++++ execute checkbox keys ++++++++++++++++++++++++++++++
			Switch(s.keycode)
				case 11:	// page up
					Key = "FullWidth"
					break
				case 12:	// page down
					Key = "MinWidth"
					break
				case 103:	// g key
					Key = "toggleSnapToGrid"
					break
				case 104:	// h key
					Key = "IntColCheckBox"
					CheckBox IntColCheckBox	,value = 1	,win=$getPanel()
					break
				case 109:	// m key
					Key = "toggleMousetoSpec"
					break
				case 110:	// n key
					Key = "toggleNormto1"
					break
				case 118:	// v key
					Key = "IntRowCheckBox"
					CheckBox IntRowCheckBox	,value = 1	,win=$getPanel()
					break
			EndSwitch
			if (StrLen(Key) > 0)
				Profile_ExecuteCheckboxAction(Key,1)
				break
			endif
			//+++++++++++++++++++++++++++++++ execute arrow keys ++++++++++++++++++++++++++++++++		
			Variable Shift = 0, Expand = 0
			Switch (Mode)																		// define a shift and expand direction (positive or negative) depending on mode and pressed key
				case 0:
					Shift	= s.keycode == 30 ? 1 : ( s.keycode == 31 ? -1 : Shift )
					Expand	= s.keycode == 29 ? 1 : ( s.keycode == 28 ? -1 : Expand )
					break
				case 1:
					Shift	= s.keycode == 29 ? 1 : ( s.keycode == 28 ? -1 : Shift )
					Expand	= s.keycode == 30 ? 1 : ( s.keycode == 31 ? -1 : Expand )
					break
			EndSwitch
			Variable Accelerate = 1																// add acceleration modifier depending on the pressed modifier key
			Accelerate = (GetKeyState(0) & 2^2) != 0 ? 4 : Accelerate
			Accelerate = (GetKeyState(0) & 2^0) != 0 ? 0.5 : Accelerate

			Variable Delta = abs(DimDelta(inwave,1 - Mode))
			Lines[1 + Mode*2] += Shift * Delta * Accelerate										// do the shift
			Lines[2 + Mode*2] += Expand * 2 * Delta * Accelerate								// do the expand
			Lines[2 + Mode*2] = Lines[2 + Mode*2] < Delta ? Delta : Lines[2 + Mode*2]			// check for the minimal width (will normally be done by the variable control)
			
			Profile_CheckBorders()																// check if the lines go over the borders of the wave
			Profile_SetFullMinCheckbox()														// toggle check-boxes
			break
		case 4:		// mouse movement
			ControlInfo/W=$getPanel() MousetoSpec
			if (V_Value)
				Variable Mx = round ((AxisValFromPixel("", "Bottom", s.mouseLoc.h)-DimOffset(inwave, 0)) / DimDelta(inwave, 0))
				Variable My = round ((AxisValFromPixel("", "Left", s.mouseLoc.v)-DimOffset(inwave, 1)) / DimDelta(inwave, 1))
				if (Mx >= DimSize(inwave, 0) || My >= DimSize(inwave, 1) || Mx < 0  || My < 0)
					break
				endif
				
				RemoveFromGraph/W=$getGraph()/Z MouseProfile
				if (Mode)																		// append one row / column under the current mouse position
					AppendToGraph/W=$getGraph()/VERT/B=bottom2 inwave[Mx][]/TN=MouseProfile
				else
					AppendToGraph/W=$getGraph()/L=left2 inwave[][My]/TN=MouseProfile
				endif
				ControlInfo/W=$getPanel() Normto1												// also normalize the mouse profile
				if (V_Value)
					if (Mode)
						Duplicate/Free/R=[Mx][] inwave, tempScale
					else
						Duplicate/Free/R=[][My] inwave, tempScale
					endif
					Variable minVal = wavemin(tempScale)
					Variable maxVal = wavemax(tempScale)
					ModifyGraph offset(MouseProfile)={,-minVal/(maxVal-minVal)} ,muloffset(MouseProfile)={,1/(maxVal-minVal)}
				endif
				ModifyGraph/W=$getGraph() rgb(MouseProfile)=(0,39321,0)
			else
				RemoveFromGraph/W=$getGraph()/Z MouseProfile
			endif

			break
	EndSwitch
	
	return HookTakeover
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function Profile_AskButton(s) : ButtonControl													// execute button commands (of the quit sub-panel)
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		Strswitch (s.ctrlName)
			case "bAskUndo":
				Profile_QuitProgram(1)
				break
			case "bAskOK":
				Profile_QuitProgram(0)
				break
			case "bAskCancel":																	// just remove the panel
				KillWindow #AskQuitPanel
				break
		EndSwitch
	endif
	return 0
End
	
//################################################################################################

Function Profile_UpdateLines()																	// generates marquee lines
	DFREF stg = root:Packages:$(kProfileWorkingDir)
	Wave LinesX	= stg:LinesX
	Wave LinesY	= stg:LinesY
	Wave Lines	= stg:CurrentLines
	
	Variable plus, minus
	if (Lines[0] == 0)																			// depending on mode switch line direction
		plus	= Lines[1] + Lines[2]/2
		minus	= Lines[1] - Lines[2]/2
		LinesX	= {-INF,INF,NaN,-INF,INF}
		LinesY	= {plus,plus,NaN,minus,minus}
	else
		plus	= Lines[3] + Lines[4]/2
		minus	= Lines[3] - Lines[4]/2
		LinesX	= {plus,plus,NaN,minus,minus}
		LinesY	= {-INF,INF,NaN,-INF,INF}
	endif

	ModifyGraph/W=$getGraph() offset(LinesY)={0,0}												// reset offset, trigger S_TraceOffset
	return 0
End

//################################################################################################

Function LinesIntegrateUpdate(Info)																// will be triggered when S_TraceOffset changes (connected via formula)
	String Info
	if( StrSearch(Info,"TNAME:LinesY;",0) <=0 )													// contain the Info line data?
		return 0
	endif
	DFREF stg = root:Packages:$(kProfileWorkingDir)
	if (DataFolderRefStatus(stg) == 0)															// something is wrong if the data is not found => exit
		return -1
	endif
	
	Wave Lines	= stg:CurrentLines
	Wave SecPr	= stg:SectionLine
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable Mode = Lines[0]
	
	if (Mode == 0)																				// depending on mode (in row [0]) read the line data (X or Y direction)
		Lines[1] += NumberByKey("YOFFSET",Info)
	else
		Lines[3] += NumberByKey("XOFFSET",Info)
	endif

	Profile_CheckBorders()																		// check if the lines go over the borders of the wave before calculating something
	//++++++++++++++++++++++++++++++++++++ prepare integration ++++++++++++++++++++++++++++++++++
	Variable Center	= Lines[1 + Mode * 2]
	Variable Width	= Lines[2 + Mode * 2]
	Variable Delta	= DimDelta(inwave, 1-Mode)
	Variable Offset	= DimOffset(inwave, 1-Mode)
	Variable Size	= DimSize(inwave, 1-Mode)

	Redimension/D/N=(DimSize(inwave, Mode)) SecPr												// set the right dimension and scale
	Variable start	= (Center -	sign(Delta) * Width/2 - Offset)/Delta + 0.5						// recalculate to rows / columns x2pnt()-like; the sign(Delta) reverses the direction for negative delta values
	Variable finish	= (Center +	sign(Delta) * Width/2 - Offset)/Delta + 0.5						// shift the values by 0.5 to get the full rows /columns
	//start	= round(start/delta*1000)*delta/1000												// round the values depending on the delta size (to get rid of the floating point errors)
	//finish= round(finish/delta*1000)*delta/1000
	start	= start < 0 ? 0 : start																// validate values
	finish	= finish > Size ? Size : finish
	//Print/D "Start initial:",start,"Finish initial:",finish									// debug

	String NoteText = ""
	NoteText += note(inwave) + "\r" + NameOfWave(inwave)+" integrated from "+num2str(ChangePrecision(Center-Width/2,4))
	NoteText += " to "+num2str(ChangePrecision(Center+Width/2,4))+" "+WaveUnits(inwave,1-Mode)+ " (width = " +num2str(Width)+" "+WaveUnits(inwave,1-Mode)+")"
	Note/K SecPr, NoteText																		// write comments
	
	Variable i, overhead = 0, rest = 0
	overhead	= ceil(start) - start															// fraction at the beginning
	start		= ceil(start)																	// beginning of the full columns / rows
	rest		= finish - trunc(finish)														// rest from the end of the range
	finish		= floor(finish)																	// maximal value
	
	if (start > finish)																			// if range is smaller than one column
		overhead += rest - 1
		rest = 0
	endif
	//Print/D "Overhead:",overhead,"Start:",start,"Finish:",finish,"Rest:",rest					// debug
	//++++++++++++++++++++++++++++++++++++ now integrate ++++++++++++++++++++++++++++++++++++++++
	SecPr = 0
	if (overhead > 0)
		SecPr += (Mode == 0 ? inwave[p][start-1]: inwave[start-1][p]) * overhead				// part of first line
	endif
	for (i = start; i < finish; i += 1)															// full lines
		SecPr += (Mode == 0 ? inwave[p][i]		: inwave[i][p])
	endfor
	if (rest > 0)
		SecPr += (Mode == 0 ? inwave[p][finish]	: inwave[finish][p]) * rest						// rest of the last line
	endif

	Width /= abs(Delta)																			// normalize for scaled waves
	if (Width > 0)
		SecPr /= Width																			// normalize to the integrated width
	endif
	//+++++++++++++++++++++++++++++++++++++++ normalize profile ++++++++++++++++++++++++++++++++++	
	ControlInfo/W=$getPanel() Normto1
	if (V_Value)																				// normalizes the spectrum to the maximum height
		start = waveMin(SecPr)
		finish = waveMax(SecPr)
		Delta = finish-start
		Delta = Delta == 0 ? 1 : Delta
		MatrixOP/O SecPr=(SecPr-start)/Delta
	endif
	SetScale/P x, DimOffset(inwave, Mode), DimDelta(inwave, Mode), WaveUnits(inwave, Mode), SecPr	// set the scale last, after MatrixOP was executed
	
	return 0
End

//################################################################################################

Function Profile_CheckBorders()																	// check if the lines are within the boundaries, if not restrict
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	
	Variable xdel = DimDelta(inwave,0), xoff = DimOffset(inwave,0)
	Variable ydel = DimDelta(inwave,1), yoff = DimOffset(inwave,1)
	Variable rows = DimSize(inwave,0) , cols = DimSize(inwave,1)
	Variable colPos	= Lines[1], colWidth= Lines[2]
	Variable rowPos	= Lines[3], rowWidth= Lines[4]
	//++++++++++++++++++++++++++ check for alignment with data grid ++++++++++++++++++++++++++++++
	Variable halfShift = 0
	ControlInfo/W=$getPanel() SnapToGrid
	if (V_Value)
		rowWidth	= xdel * round(rowWidth/xdel)
		colWidth	= ydel * round(colWidth/ydel)
		halfShift	= xdel/2*!mod(rowWidth/xdel,2)												// for even widths the center needs to be shifted over by half a row/col
		rowPos		= xoff + xdel*round((rowPos - halfShift - xoff)/xdel) + halfShift
		halfShift	= ydel/2*!mod(colWidth/ydel,2)
		colPos		= yoff + ydel*round((colPos - halfShift - yoff)/ydel) + halfShift
	endif
	//++++++++++++++++++++++++++++ read and check y position data ++++++++++++++++++++++++++++++++			
	Variable Boundary1 = yoff + (colWidth*sign(ydel) - ydel)/2
	Variable Boundary2 = yoff - (colWidth*sign(ydel) - ydel)/2 + ydel*(cols-1)
	
	colPos = (colPos - Boundary1)*sign(ydel) < 0 ? Boundary1 : colPos							// too low?
	colPos = (colPos - Boundary2)*sign(ydel) > 0 ? Boundary2 : colPos							// too high?
	Lines[1] = colPos																			// return correct value
	colWidth = colWidth > abs(ydel*cols) ? abs(ydel*cols) : colWidth							// too big?
	Lines[2] = colWidth
	//++++++++++++++++++++++++++++ read and check x position data ++++++++++++++++++++++++++++++++
	Boundary1 = xoff + (rowWidth*sign(xdel) - xdel)/2
	Boundary2 = xoff - (rowWidth*sign(xdel) - xdel)/2 + xdel*(rows-1)
	
	rowPos = (rowPos - Boundary1)*sign(xdel) < 0 ? Boundary1 : rowPos							// left check
	rowPos = (rowPos - Boundary2)*sign(xdel) > 0 ? Boundary2 : rowPos							// right check
	Lines[3] = rowPos
	rowWidth = rowWidth > abs(xdel*rows) ? abs(xdel*rows) : rowWidth							// too big?
	Lines[4] = rowWidth
	
	Profile_UpdateLines()																		// set the lines to the restricted values
	return 0
End

//#################################support functions for control elements#########################

Function ModeCheckBoxControl(s) : CheckBoxControl
	STRUCT WMCheckboxAction &s
	if(s.eventCode == 2)
		Profile_ExecuteCheckboxAction(s.ctrlName,s.checked)
	endif
	return 0
End
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function Profile_ExecuteCheckboxAction(type,checked)
	String type
	Variable checked
	
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	StrSwitch (type)
		case "IntColCheckBox":
			Lines[0] = 0
			Profile_AppendSectionTrace()
			CheckBox IntRowCheckBox ,win=$getPanel() ,value = 0
		break
		case "IntRowCheckBox":
			Lines[0] = 1
			Profile_AppendSectionTrace()
			CheckBox IntColCheckBox ,win=$getPanel() ,value = 0
		break
		case "FullWidth":
			Profile_SetMinFullLineValues(1,Lines[0],checked)
		break
		case "MinWidth":
			Profile_SetMinFullLineValues(0,Lines[0],checked)
		break
		case "toggleNormto1":
			ControlInfo/W=$getPanel() Normto1
			CheckBox Normto1 ,win=$getPanel() ,value = 1-V_Value
		break
		case "toggleMousetoSpec":
			ControlInfo/W=$getPanel() MousetoSpec
			CheckBox MousetoSpec ,win=$getPanel() ,value = 1-V_Value
		break
		case "toggleSnapToGrid":
			ControlInfo/W=$getPanel() SnapToGrid
			CheckBox SnapToGrid  ,win=$getPanel() ,value = 1-V_Value
		break
	EndSwitch
	
	Profile_CheckBorders()																		// check if the lines go over the borders of the wave
	Profile_SetFullMinCheckbox()																// toggle check-boxes
	return 0
End

//################################################################################################

Function Profile_SetMinFullLineValues(type,mode,activate)	
	Variable type, mode, activate																// type 0: min width; type 1: full width

	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable size	= DimSize(inwave,1-mode)
	Variable offset	= DimOffset(inwave,1-mode)
	Variable delta	= DimDelta(inwave,1-mode)
	
	Variable init = abs(delta)*round((size-1) / 20)												// calculate ca. 5% of width
	init = init == 0 ? abs(delta) : init														// make sure that's not zero
	if (type)
		Lines[1 + mode*2]	= offset + delta*(size-1)/2											// center position
		Lines[2 + mode*2]	= activate == 1 ? abs(delta)*size : init							// full with or ca. 20 %
	else
		Lines[2 + mode*2]	= activate == 1 ? abs(delta) : init									// one line or 20 %
	endif
	return 0
End

//################################################################################################

Function ChangeParameterControl(s) : SetVariableControl											// execute variable changes and sets the checkboxes
	STRUCT WMSetVariableAction &s
	if (s.eventCode == 1 || s.eventCode == 2)
		Profile_CheckBorders()																	// check if the lines go over the borders of the wave
		Profile_SetFullMinCheckbox()															// set check-boxes correctly
	endif
	return 0
End
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function Profile_SetFullMinCheckbox()
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable Mode = Lines[0]
	Variable Delta	= abs(DimDelta(inwave,1 - Mode))											// look, if a checkbox could be applied
	Variable Count	= DimSize(inwave,1 - Mode)
	CheckBox FullWidth	,win=$getPanel()	,value = Lines[2+Mode*2] >= Delta * Count			// check for full width
	CheckBox MinWidth	,win=$getPanel()	,value = Lines[2+Mode*2] <= Delta					// check for minimal width
	return 0
End

//################################################################################################

Function Profile_ContrastSliderProc(s) : SliderControl
	STRUCT WMSliderAction &s
	if (s.eventCode & 1)
		ControlInfo/W=$getPanel() ContrastLow
		Variable minVal = V_Value
		ControlInfo/W=$getPanel() ContrastHigh
		Variable maxVal = V_Value
		ControlInfo/W=$getPanel() ContrastColor
		String color = S_Value
		ModifyImage/W=$getGraph() ''#0 ctab={minVal,maxVal,$color,0}
	endif
	return 0
End

//################################################################################################

Function Profile_ContrastColorProc(s) : PopupMenuControl
	STRUCT WMPopupAction &s
	if (s.eventCode == 2)
		ControlInfo/W=$getPanel() ContrastLow
		Variable minVal = V_Value
		ControlInfo/W=$getPanel() ContrastHigh
		Variable maxVal = V_Value
		ModifyImage/W=$getGraph() ''#0 ctab={minVal,maxVal,$s.popStr,0}
	endif
	return 0
End

//################################################################################################

Function Profile_MainButtonProc(s) : ButtonControl
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		Profile_ExecuteButtonCommand(s.ctrlName)
	endif
	return 0
End
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function Profile_ExecuteButtonCommand(which)
	String which

	DFREF stg = root:Packages:$(kProfileWorkingDir)
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable i, CurSum = 0, shiftDir = 1
	Variable hasChanged = 0																		// updates the contrast sliders (1 = don't change range, 2 = reset range)
	Variable Mode = Lines[0]
	
	String logStr = ""
	String from	= num2str(ChangePrecision(Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2,4))		// current range
	String to	= num2str(ChangePrecision(Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2,4))
	
	Strswitch (which)
		case "bSaveQuit":
			Profile_QuitProgram(0)
			return 0
		break
		case "bSaveSpec":
			ProfileAction_SaveProfile()
		break
		case "bCompare":
			ProfileAction_Compare()
		break
		case "bUndoAll":
			Profile_UndoLastUpdate("all")
			logStr = NameOfWave(inwave) + " (" + time() + "): All changes undone."
			hasChanged = 2
		break
		case "bUndoLast":
			Profile_UndoLastUpdate("last")
			logStr = NameOfWave(inwave) + " (" + time() + "): Last change undone."
			hasChanged = 2
		break
		case "bAlign":
			logStr = ProfileAction_Align()
			hasChanged = 1
		break
		case "bNormalize":																		// normalize to profile shape
			Profile_UndoLastUpdate("save")
			if( !((GetkeyState(0) & 2^0) != 0) && kNormProfileSmooth > 0)						// ctrl / command = no smoothing
				Smooth round(kNormProfileSmooth), SecPr											// smooth the profile a bit
			endif
			for (i = 0; i < DimSize(inwave,1-Mode); i += 1)
				if (Mode)
					inwave[i][] /= SecPr[q]
				else
					inwave[][i] /= SecPr[p]	
				endif
			endfor
			CurSum = mean(SecPr)
			
			logStr = NameOfWave(inwave) + " (" + time() + "): Normalized to "+SelectString((GetkeyState(0) & 2^0),"","(exact) ")
			logStr += SelectString(Mode,"horizontal","vertical")+" profile (Range: "+from+"-"+to+")."
			hasChanged = 2
		break
		case "bNormMax":																		// normalize to maximum
			Profile_UndoLastUpdate("save")
			Make/FREE/D/N=(DimSize(inwave,1-Mode)) tempwave										// get a temporary wave
			for (i = 0; i < DimSize(inwave,Mode); i += 1)
				tempwave = Mode > 0 ? inwave[p][i] : inwave[i][p]								// extract the row / column
				Variable maxV = Wavemax(tempwave)
				maxV = maxV == 0 ? 1 : maxV
				if (Mode)
					inwave[][i] /= maxV
				else
					inwave[i][] /= maxV
				endif
				CurSum += maxV																	// add the values for re-normalization later
			endfor
			CurSum /= i+1
			logStr = NameOfWave(inwave) + " (" + time() + "): Normalized to each "+SelectString(Mode,"vertical","horizontal")+" line maximum."
			hasChanged = 2
		break
		case "bNormUser":																		// normalize to measurement value
			Profile_UndoLastUpdate("save")
			CurSum = Profile_ExecuteUserInfoButton(inwave)
			Button bNormUser ,win=$getPanel() ,disable= 2										// disable the button to avoid double clicking (more than once is meaningless anyway)			
			logStr = NameOfWave(inwave) + " (" + time() + "): Normalized by "+kUserInfoBtnName+"."
			hasChanged = 2
		break
		case "bSubtractPr":		
			Profile_UndoLastUpdate("save")
			ProfileAction_SubtractProfile()
			logStr = NameOfWave(inwave) + " (" + time() + "): "+SelectString(Mode,"Horizontal","Vertical")+" profile (Range: "+from+"-"+to+") subtracted."
			hasChanged = 1
		break
		case "bRemoveSelection":
			Profile_UndoLastUpdate("save")
			logStr = ProfileAction_RemoveSelection()
			hasChanged = 1
		break
		case "bShiftYAxisMinus":
			shiftDir = -1
		case "bShiftYAxisPlus":
			Profile_UndoLastUpdate("save")
			Profile_ShiftButtonStatus(shiftDir)													// enable and disable shift buttons
			ProfileAction_ShiftImage(shiftDir)
			logStr = NameOfWave(inwave) + " (" + time() + "): Data shifted to "+SelectString(shiftDir+1,"negative","positive")+" y-axis values."
			hasChanged = 1
		break
	Endswitch
	
	ControlInfo/W=$getPanel() MaintainInt
	if (V_Value && CurSum != 0 && numtype(CurSum) == 0)											// restore the average height of the spectrum (from the average height of the Profile)
		inwave *= CurSum
		logStr += " Average intensity is maintained."
	endif
	
	ControlInfo/W=$getPanel() cPrintEdits														// print logs if desired
	if (V_flag == 2 && V_Value == 1 && strlen(logStr))
		Print logStr
	endif
	
	ControlInfo/W=$getPanel() ContrastColor
	if (hasChanged > 0 && abs(V_flag) == 3)														// update the contrast sliders
		String color = S_Value
		Variable minVal, maxVal
		
		if (hasChanged == 1)
			ControlInfo/W=$getPanel() ContrastLow
			minVal = limit(V_Value,WaveMin(inwave),WaveMax(inwave))
			ControlInfo/W=$getPanel() ContrastHigh
			maxVal = limit(V_Value,WaveMin(inwave),WaveMax(inwave))
			minVal = minVal >= maxVal ? WaveMin(inwave) : minVal
		else
			minVal = WaveMin(inwave)
			maxVal = WaveMax(inwave)
		endif

		Slider ContrastLow	,win=$getPanel()	,value=minVal	,limits={WaveMin(inwave),WaveMax(inwave),0}
		Slider ContrastHigh	,win=$getPanel()	,value=maxVal	,limits={WaveMin(inwave),WaveMax(inwave),0}
		if (minVal == WaveMin(inwave) && maxVal != WaveMax(inwave))
			ModifyImage/W=$getGraph() ''#0 ctab={*,maxVal,$color,0}
		elseif (minVal != WaveMin(inwave) && maxVal == WaveMax(inwave))
			ModifyImage/W=$getGraph() ''#0 ctab={minVal,*,$color,0}
		elseif (minVal == WaveMin(inwave) && maxVal == WaveMax(inwave))
			ModifyImage/W=$getGraph() ''#0 ctab={*,*,$color,0}
		else
			ModifyImage/W=$getGraph() ''#0 ctab={minVal,maxVal,$color,0}
		endif
	endif

	Profile_UpdateLines()																		// update
	return 0
End

//################################################################################################

Function ProfileAction_SaveProfile()
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	Wave SecPr	= stg:SectionLine
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable Mode	= Lines[0]
	
	ControlInfo/W=$getPanel() FullWidth
	Variable isFull  = V_Value
	Variable skipEnd = (GetkeyState(0) & 2^2) != 0												// enables a keyboard shortcut to just cut the name from the ending
	
	String name = NameOfWave(inwave)
	String outName
	if (isFull || skipEnd)
		outName	= name+"_pr"+SelectString(Mode , "H", "V")
	else
		String from	= num2str(ChangePrecision(Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2,4))	// current range
		String to	= num2str(ChangePrecision(Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2,4))
		Variable chars = inf
		if (IgorVersion() < 7.00 && (strlen(name) + strlen(from) + strlen(to) + 5) > 31 )		// make sure to not go over the 31 character limit
			from	= from[0,5]
			to		= to[0,5]
			chars	= 25 - strlen(from) - strlen(to)
		endif
		outName = name[0,chars] + SelectString(Mode, "_H(" , "_V(") + from+"-"+to + ")"
	endif
	Duplicate/O SecPr, $outName
	return 0
End
//++++++++++++++++++++++++++++++++++++++++ number rounding +++++++++++++++++++++++++++++++++++++++
Static Function ChangePrecision(value, numSigDigits)											// round values to select-able number of digits
	variable value, numSigDigits
	string str
	sprintf str, "%.*g\r", numSigDigits, value
	return str2num(str)
End

//################################################################################################

Function ProfileAction_Compare()
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	RemoveFromGraph/W=$getGraph()/Z CompareLine
	Duplicate/O SecPr, stg:CompareLine															// duplicate current profile for comparison
	if (Lines[0] == 1)																			// append together with the main profile
		AppendToGraph/W=$getGraph()/VERT/B=bottom2 stg:CompareLine
	else
		AppendToGraph/W=$getGraph()/L=left2 stg:CompareLine
	endif
	ModifyGraph/W=$getGraph() rgb(CompareLine)=(0,0,52224)
	return 0
End

//################################################################################################

Function/S ProfileAction_RemoveSelection()
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	
	Variable Mode	= Lines[0]
	Variable from	= Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2								// calculate the position of the Lines
	Variable to		= Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2
	Variable offset = DimOffset(inwave,1-Mode), delta = DimDelta(inwave,1-Mode)
	Variable nOff	= offset-(from-to)															// recalculate the offset (is used if the removed section is at the left / bottom side)
	//Print "start:",from,"stop:", to, "offset:", nOff											// debug (range in wave scales)
	from	= round((from + delta/2 - offset)/delta)											// which points do the lines include
	to		= round((to + delta/2 - offset)/delta)
	//Print "start:",from,"stop:", to, "count:", abs(to - from)									// debug (range in wave points)
	
	DeletePoints/M=(1-Mode) from, abs(to-from), Inwave											// delete the selected points
	if (from == 0)																				// shifts the scale only if the wave is cut at the old offset
		if (Mode)
			SetScale/P x nOff, delta, WaveUnits(inwave,1-Mode),  inwave
		else
			SetScale/P y nOff, delta, WaveUnits(inwave,1-Mode),  inwave
		endif
	endif
	
	String logStr = NameOfWave(inwave) + " (" + time() + "): "
	logStr += SelectString(Mode,"Horizontal","Vertical")+" ROI section deleted ("
	logStr += num2str(abs(to-from))+" lines in point range "+num2str(from)+"-"+num2str(to)+")."
	return logStr
End

//################################################################################################

Function ProfileAction_ShiftImage(shiftDir)
	Variable shiftDir
	
	DFREF stg = root:Packages:$(kProfileWorkingDir)
	SVAR WLoc = stg:InwaveLocation
	Wave inwave = $WLoc
	Wave/Z lastcpy = stg:LastCopy																// read wave data from the copy if available
	if (!WaveExists(lastcpy))
		Wave lastcpy = $WLoc
	endif
	
	Variable modkey	= (GetKeyState(0) & 2^0) != 0												// when the 'ctrl/command' key is pressed, the excess amount will be cut from the front of the spectrum	
	Variable rows = DimSize(lastcpy,0),  cols = DimSize(lastcpy,1)
	Variable xdel = DimDelta(lastcpy,0), xoff = DimOffset(lastcpy,0)
	Variable ydel = DimDelta(lastcpy,1), yoff = DimOffset(lastcpy,1)
	
	Variable addrow = ceil((cols-1)*abs(ydel/xdel))												// how many rows need to be added for the shift
	Variable nOff	= xoff + shiftDir*yoff														// the new x-offset
	nOff += sign(ydel)<0 ? shiftDir*(cols-1)*ydel : 0											// for negative y-deltas add the full width (basically start from the bottom)
	if (modkey && (addrow < rows-10))															// delete incomplete parts
		Redimension/D/N=(rows-addrow,-1) inwave
		nOff += shiftDir ==  1*sign(xdel) ? addrow*xdel : 0
	else
		Redimension/D/N=(rows+addrow,-1) inwave
		nOff -= shiftDir == -1*sign(xdel) ? sign(xdel)*(cols-1)*abs(ydel) : 0					// for negative shifts the start offset needs to be shifted as well (by the y-width of the spectrum)
	endif
	SetScale/P x nOff, xdel, WaveUnits(inwave,0), inwave

	inwave[][] = interp2D(lastcpy, nOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel)			// calculate every shifted point
	MatrixOP/O inwave=replaceNaNs(inwave,0)														// fill NaNs with zero
	//+++++++++++++++++++++++ if there are zero rows, remove them ++++++++++++++++++++++++
	Variable i
	for (i = 0; i < DimSize(inwave,0); i += 1)
		MatrixOp/Free partSum = Sum(Row(inwave,i))
		if (partSum[0] != 0)
			break
		endif
	endfor
	if (i > 0)
		DeletePoints 0,i, inwave
		SetScale/P x DimOffset(inwave,0)+i*DimDelta(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0),  inwave
	endif
	
	Variable excessRows = 0, lastData = -1
	for (i = DimSize(inwave,0)-1; i > 0; i -= 1)
		MatrixOp/Free partSum = Sum(Row(inwave,i))
		if (partSum[0] == 0)
			excessRows += 1
		else
			if (lastData == -1)
				lastData = i
			endif
		endif
	endfor
	if (excessRows > 0 && lastData != -1)
		DeletePoints lastData+1, excessRows, inwave
	endif
	
	return 0
End

//################################################################################################

Function Profile_ShiftButtonStatus(mode)														// disables the button after one press to prevent multiple shifts
	Variable mode			// -1 = -Y; 0 = undo; 1 = +Y;
	Variable currMode = 0																		// first, find the current disable state of the two buttons
	ControlInfo/W=$getPanel() bShiftYAxisPlus
	currMode = (V_disable == 2) ?  1 : currMode
	ControlInfo/W=$getPanel() bShiftYAxisMinus
	currMode = (V_disable == 2) ? -1 : currMode
	Variable Yplus  = currMode == 0 && mode ==  1 ? 2 : 0										// deactivate the relevant button
	Variable Yminus = currMode == 0 && mode == -1 ? 2 : 0
	Button bShiftYAxisPlus  ,win=$getPanel() ,disable=Yplus
	Button bShiftYAxisMinus ,win=$getPanel() ,disable=Yminus
	Button bUndoLast		,win=$getPanel() ,userdata(lastShift)=num2str(mode)
	return 0
End

//################################################################################################

Function ProfileAction_SubtractProfile()
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	
	Variable Mode = Lines[0]
	Variable xOff = DimOffset(inwave,0), xDel = DimDelta(inwave,0)
	Variable yOff = DimOffset(inwave,1), yDel = DimDelta(inwave,1)
	Variable size = DimSize(inwave,1-Mode), i
	
	Variable LeftOrRight = 0
	LeftOrRight = GetKeyState(0) == 1 ?  1 : LeftOrRight										// ctrl/command for positive shift
	LeftOrRight = GetKeyState(0) == 2 ? -1 : LeftOrRight										// alt/option for negative shift
	if (Mode == 0 && LeftOrRight != 0)															// subtracts along the Y axis (increasing/decreasing with Y values)
		Variable xadd  = ceil(size*abs(yDel/xDel) )												// Add points at both sides to stay within bounds
		Variable xsize = DimSize(inwave,0)
		Variable nOff  = xOff-xDel*xadd - LeftOrRight*(Lines[1]-yOff)							// new offset: add extended range and offset from current Y-center (Lines[1])
		Make/D/Free/N=(xsize+2*xadd) tempcut
		SetScale/P x, nOff, xDel ,tempcut														// readjust scale for additional points
		tempcut[0,xadd-1]					= SecPr[0]											// fill with extended profile (endpoints will be repeated)
		tempcut[xadd,xsize+xadd-1]			= SecPr[p-xadd]
		tempcut[xsize+xadd,xsize+2*xadd-1]	= SecPr[xsize-1]
	endif

	for (i = 0; i < size; i += 1)
		if (Mode == 1)
			inwave[i][] -= SecPr[q]
		else
			if(WaveExists(Tempcut))
				inwave[][i] -= tempcut(x-LeftOrRight*yDel*i)									// subtract a moving profile
			else
				inwave[][i] -= SecPr[p]															// subtract stationary profile
			endif
		endif
	endfor
	return 0
End

//################################################################################################

Function/S ProfileAction_Align()
	DFREF stg	= root:Packages:$(kProfileWorkingDir)
	Wave Lines	= stg:CurrentLines
	SVAR WLoc	= stg:InwaveLocation
	Wave inwave	= $WLoc
	Variable Mode = Lines[0]
	
	Wave/Z backup = stg:TempCopy																// create an extra save of the current state
	if (!WaveExists(backup))
		Duplicate inwave, stg:TempCopy
		Wave backup = stg:TempCopy
	endif
	
	String name 	= NameOfWave(inwave)
	Variable size	= DimSize(inwave,1-Mode)
	Variable line	= DimSize(inwave,Mode)
	
	ControlInfo/W=$getPanel() cAlignSmooth
	Variable smth = V_Value == 1
	ControlInfo/W=$getPanel() cAlignOutput
	Variable output = V_Value == 1
	ControlInfo/W=$getPanel() SmoothFactor
	Variable smthFact = abs(V_Flag) == 5 ? V_Value : ceil(line*kAlignSmooth/100)
	Variable rev = (GetkeyState(0) & 2^2) != 0 ? 1 : 0											// shift = reverse sorting 0 = no, 1 = yes

	String logStr = name + " (" + time() + "): Aligned to the "
	
	Variable chars = IgorVersion() < 7.00 ? 25: inf
	if (output)																					// writes align shift into a file
		Make/O/D/N=(size) $(name[0,chars]+"_shft")
		Wave out = $(name[0,chars]+"_shft")
		SetScale/P x DimOffset(inwave,1-Mode), DimDelta(inwave,1-Mode), WaveUnits(inwave,1-Mode), out	
	endif

	Variable from = 0, to = line-1, count														// use the cursors to select a sub-region for alignment
	if (strlen(CsrInfo(A)) > 0)
		from = Mode ? qcsr(A,"") : pcsr(A,"")
		from = numtype(from)	== 0 ? from : 0
	endif
	if (strlen(CsrInfo(B)) > 0)
		to	 = Mode ? qcsr(B,"") : pcsr(B,"")
		to	 = numtype(to)	== 0 ? to : line-1
	endif
	if (from > to)
		count = from
		from  = to
		to	  = count
	endif
	
	Make/FREE/N=(line) ShiftWorkA = 0, ShiftWorkB = 0
	if (Mode)
		ShiftWorkA[from,to] = backup[0 + rev*(size-1)][p]
		logStr += SelectString(rev,"left","right") + " side"
	else
		ShiftWorkA[from,to] = backup[p][0 + rev*(size-1)]
		logStr += SelectString(rev,"bottom","top")
	endif
	if(smth && smthFact > 0)
		Smooth smthFact, ShiftWorkA
		logStr += " with smoothing (factor " + num2str(smthFact) + ")"
	endif
	if (output)
		logStr += " => output to " + NameOfWave(out)
	endif
	
	inwave = NaN
	for (count = 0; count < size; count += 1)													// find the correlation and shift for each row/column
		if (Mode)
			ShiftWorkB[from,to] = backup[count][p]
		else
			ShiftWorkB[from,to] = backup[p][count]
		endif
		if(smth && smthFact > 0)
			Smooth smthFact, ShiftWorkB
		endif
		
		Variable align = SpectrumCorrelationShift(ShiftWorkA,ShiftWorkB)
		if (Mode)
			inwave[count][-align*(align < 0),line-1 -align*(align >= 0)] = backup[count][q+align]
		else
			inwave[-align*(align < 0),line-1 -align*(align >= 0)][count] = backup[p+align][count]
		endif
		
		if(output)
			out[count] = DimDelta(inwave, Mode) * align
		endif
	endfor
	return logStr+"."
End

//+++++++++++++++++++++++++++++++++++++ correlation function +++++++++++++++++++++++++++++++++++++

static Function SpectrumCorrelationShift(in1,in2)												// Helper function: Finds maximum correlation of two slices
	wave in1,in2																				// data to compare
	Duplicate/Free in2, work
	Correlate/AUTO in1, work
	WaveStats/Q work
	if (V_sum != 0)
		return V_maxRowLoc - x2pnt(work, 0)
	else		// perfect match
		return 0
	endif
End

//################################################################################################

Function Profile_UndoLastUpdate(which)
	String which

	DFREF stg = root:Packages:$(kProfileWorkingDir)
	Wave startcopy	= stg:StartCopy
	Wave/Z lastcopy = stg:LastCopy
	Wave/Z tempcopy	= stg:TempCopy
	SVAR WLoc		= stg:InwaveLocation
	Wave inwave		= $WLoc

	StrSwitch(which)
		case "save":
			Duplicate/O inwave, stg:LastCopy
		break
		case "last":
			if (WaveExists(tempcopy))
				Duplicate/O tempcopy, inwave, lastcopy											// updating lastcopy will prevent loading an older version upon second press
			elseif (WaveExists(lastcopy))
				Duplicate/O lastcopy, inwave
			else
				Duplicate/O startcopy, inwave
			endif
			
			String last = GetUserData("ProfileGeneratorWindow","bUndoLast","lastShift")			// checks if shift button should be restored
			if (strlen(last) > 0)
				Profile_ShiftButtonStatus(-str2num(last))
			endif
		break
		case "all":
			Duplicate/O startcopy, inwave
			KillWaves/Z lastcopy
			Profile_ShiftButtonStatus(0)
		break
	EndSwitch
	KillWaves/Z tempcopy
	
	Button bUndoLast ,win=$getPanel() ,userdata(lastShift)=""									// delete shift-button info
	Button bNormUser ,win=$getPanel() ,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)	// re-enable the button if data is available
	return 0
End